
#include "ShapeSphere.h" // check
#include "ShapeConvexPolygon.h" // check
#include "ShapeSpheropolygon.h" // check
#include "ShapePolyhedron.h"
#include "ShapeConvexPolyhedron.h" // check
#include "ShapeSpheropolyhedron.h" // check
#include "ShapeSimplePolygon.h" // check
#include "ShapeEllipsoid.h" // check
#include "ShapeFacetedSphere.h"
#include "ShapeSphinx.h"
#include "ShapeUnion.h"
#include "hoomd/extern/gsd.h"
#include "hoomd/GSDDumpWriter.h"
#include "hoomd/GSDReader.h"
#include "hoomd/HOOMDMPI.h"

#include <string>
#include <memory>
#include <stdexcept>
#include <algorithm>
#include <numeric>

#ifndef _GSD_HPMC_Schema_H_
#define _GSD_HPMC_Schema_H_

template<class T>
using param_array = typename std::vector<T, managed_allocator<T> >;

struct gsd_schema_hpmc_base
    {
    gsd_schema_hpmc_base(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : m_exec_conf(exec_conf), m_mpi(mpi) {}
    const std::shared_ptr<const ExecutionConfiguration> m_exec_conf;
    bool m_mpi;
    };

struct gsd_schema_hpmc : public gsd_schema_hpmc_base
    {
    gsd_schema_hpmc(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}
    template<class T>
    int write(gsd_handle& handle, const std::string& name, unsigned int Ntypes, const T* const data, gsd_type type)
        {
        if(!m_exec_conf->isRoot())
            return 0;
        int retval = 0;
        retval |= gsd_write_chunk(&handle, name.c_str(), type, Ntypes, 1, 0, (void *)data);
        return retval;
        }

    template<class T>
    bool read(std::shared_ptr<GSDReader> reader, uint64_t frame, const std::string& name, unsigned int Ntypes, T* const data, gsd_type type)
        {
        bool success = true;
        std::vector<T> d;
        if(m_exec_conf->isRoot())
            {
            d.resize(Ntypes);
            success = reader->readChunk((void *) &d[0], frame, name.c_str(), Ntypes*gsd_sizeof_type(type), Ntypes) && success;
            }
    #ifdef ENABLE_MPI
        if(m_mpi)
            {
            bcast(d, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
            }
    #endif
        if(!d.size())
            throw std::runtime_error("Error occured while attempting to restore from gsd file.");
        for(unsigned int i = 0; i < Ntypes; i++)
            {
            data[i] = d[i];
            }
        return success;
        }
    };

template<class T>
struct gsd_shape_schema : public gsd_schema_hpmc_base
    {
    gsd_shape_schema(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}

    int write(gsd_handle&, const std::string&, unsigned int, const param_array<T>& )
        {
        throw std::runtime_error("This is not implemented");
        return 0;
        }
    bool read(std::shared_ptr<GSDReader>, uint64_t, const std::string&, unsigned int, param_array<T>&)
        {
        throw std::runtime_error("This is not implemented");
        return false;
        }
    };

template<>
struct gsd_shape_schema<hpmc::sph_params>: public gsd_schema_hpmc_base
    {
    gsd_shape_schema(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}

    int write(gsd_handle& handle, const std::string& name, unsigned int Ntypes, const param_array<hpmc::sph_params>& shape)
        {
        if(!m_exec_conf->isRoot())
            return 0;
        int retval = 0;
        std::string path = name + "radius";
        std::vector<float> data(Ntypes);
        std::transform(shape.begin(), shape.end(), data.begin(), [](const hpmc::sph_params& s)->float{return s.radius;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&data[0]);
        return retval;
        }

    bool read(  std::shared_ptr<GSDReader> reader,
                uint64_t frame,
                const std::string& name,
                unsigned int Ntypes,
                param_array<hpmc::sph_params>& shape
            )
        {
        bool success = true;
        std::string path = name + "radius";
        std::vector<float> data;
        if(m_exec_conf->isRoot()){
            data.resize(Ntypes, 0.0);
            success = reader->readChunk((void *) &data[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
        }

    #ifdef ENABLE_MPI
        if(m_mpi)
            {
            bcast(data, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
            }
    #endif
        if(!data.size()) // adding this sanity check but can remove.
            throw std::runtime_error("Error occured while attempting to restore from gsd file.");
        for(unsigned int i = 0; i < Ntypes; i++)
            {
            shape[i].radius = data[i];
            }
        return success;
        }
    };

template<>
struct gsd_shape_schema<hpmc::ell_params>: public gsd_schema_hpmc_base
    {
    gsd_shape_schema(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}

    int write(gsd_handle& handle, const std::string& name, unsigned int Ntypes,const param_array<hpmc::ell_params>& shape)
        {
        if(!m_exec_conf->isRoot())
            return 0;

        int retval = 0;
        std::vector<float> data(Ntypes);
        std::string path = name + "a";
        std::transform(shape.cbegin(), shape.cend(), data.begin(), [](const hpmc::ell_params& s)->float{return s.x;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&data[0]);
        path = name + "b";
        std::transform(shape.cbegin(), shape.cend(), data.begin(), [](const hpmc::ell_params& s)->float{return s.y;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&data[0]);
        path = name + "c";
        std::transform(shape.cbegin(), shape.cend(), data.begin(), [](const hpmc::ell_params& s)->float{return s.z;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&data[0]);
        return retval;
        }

    bool read(  std::shared_ptr<GSDReader> reader,
                uint64_t frame,
                const std::string& name,
                unsigned int Ntypes,
                param_array<hpmc::ell_params>& shape
            )
        {
        bool success = true;
        std::vector<float> a,b,c;
        if(m_exec_conf->isRoot())
            {
            a.resize(Ntypes),b.resize(Ntypes),c.resize(Ntypes);
            std::string path = name + "a";
            success = reader->readChunk((void *)&a[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
            path = name + "b";
            success = reader->readChunk((void *)&b[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
            path = name + "c";
            success = reader->readChunk((void *)&c[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
            }
        #ifdef ENABLE_MPI
            if(m_mpi)
                {
                bcast(a, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                bcast(b, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                bcast(c, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                }
        #endif

        if(!a.size() || !b.size() || !c.size()) // adding this sanity check but can remove.
            throw std::runtime_error("Error occured while attempting to restore from gsd file.");

        for(unsigned int i = 0; i < Ntypes; i++)
            {
            shape[i].x = a[i];
            shape[i].y = b[i];
            shape[i].z = c[i];
            }
        return success;
        }
    };

template<>
struct gsd_shape_schema< hpmc::detail::poly3d_verts > : public gsd_schema_hpmc_base
    {
    gsd_shape_schema(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}

    int write(gsd_handle& handle, const std::string& name, unsigned int Ntypes, const param_array<hpmc::detail::poly3d_verts>& shape)
        {
        if(!m_exec_conf->isRoot())
            return 0;

        std::string path;
        int retval = 0;
        std::vector<uint32_t> N(Ntypes);
        path = name + "N";
        std::transform(shape.cbegin(), shape.cend(), N.begin(), [](const hpmc::detail::poly3d_verts& s) -> uint32_t{return s.N;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_UINT32, Ntypes, 1, 0, (void *)&N[0]);
        path = name + "vertices";
        size_t count = std::accumulate(N.begin(), N.end(), 0);
        std::vector<float> data(count*Ntypes*3), sr(Ntypes);
        count = 0;
        for(unsigned int i = 0; i < Ntypes; i++)
            {
            for (unsigned int v = 0; v < shape[i].N; v++)
                {
                data[v*3+0+count] = float(shape[i].x[v]);
                data[v*3+1+count] = float(shape[i].y[v]);
                data[v*3+2+count] = float(shape[i].z[v]);
                }
            count += shape[i].N;
            }
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, count, 3, 0, (void *)&data[0]);
        path = name + "sweep_radius";
        std::transform(shape.cbegin(), shape.cend(), sr.begin(), [](const hpmc::detail::poly3d_verts& s) -> float{return s.sweep_radius;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&sr[0]);
        return retval;
        }

    bool read(  std::shared_ptr<GSDReader> reader,
                uint64_t frame,
                const std::string& name,
                unsigned int Ntypes,
                param_array<hpmc::detail::poly3d_verts>& shape
            )
        {
        bool success = true;
        std::vector<float> vertices,sweep_radius;
        std::vector<uint32_t> N;
        uint32_t count = 0;
        std::string path;
        assert(shape.size() == Ntypes);
        if(m_exec_conf->isRoot())
            {
            N.resize(Ntypes);
            sweep_radius.resize(Ntypes);
            path = name + "N";
            success = reader->readChunk((void *)&N[0], frame, path.c_str(),  Ntypes*gsd_sizeof_type(GSD_TYPE_UINT32), Ntypes) && success;
            count = std::accumulate(N.begin(), N.end(), 0);
            vertices.resize(count*Ntypes*3);
            path = name + "vertices";
            success = reader->readChunk((void *)&vertices[0], frame, path.c_str(), 3*count*gsd_sizeof_type(GSD_TYPE_FLOAT), count) && success;
            path = name + "sweep_radius";
            success = reader->readChunk((void *)&sweep_radius[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
            }
    #ifdef ENABLE_MPI
        if(m_mpi)
            {
            bcast(N, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
            bcast(vertices, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
            bcast(sweep_radius, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
            }
    #endif
        if(!N.size() || !vertices.size() || !sweep_radius.size()) // adding this sanity check but can remove.
            throw std::runtime_error("Error occured while attempting to restore from gsd file.");

        count = 0;
        for (unsigned int i = 0; i < Ntypes; i++)
            {
            float dsq = 0.0;
            hpmc::detail::poly3d_verts result(N[i], m_exec_conf->isCUDAEnabled());
            for (unsigned int v = 0; v < N[i]; v++)
                {
                result.x[v] = vertices[v*3+0+count];
                result.y[v] = vertices[v*3+1+count];
                result.z[v] = vertices[v*3+2+count];
                dsq = fmax(result.x[v]*result.x[v] + result.y[v]*result.y[v] + result.z[v]*result.z[v], dsq);
                }
            result.diameter = 2.0*(sqrt(dsq)+result.sweep_radius);
            result.N = N[i];
            result.sweep_radius = sweep_radius[i];
            count += N[i];
            shape[i] = result; // Can we avoid a full copy of the data (move semanitcs?)
            }
        return success;
        }
    };

template<>
struct gsd_shape_schema< hpmc::detail::poly2d_verts >: public gsd_schema_hpmc_base
    {
    gsd_shape_schema(const std::shared_ptr<const ExecutionConfiguration> exec_conf, bool mpi) : gsd_schema_hpmc_base(exec_conf, mpi) {}

    int write(gsd_handle& handle, const std::string& name, unsigned int Ntypes, const param_array<hpmc::detail::poly2d_verts>& shape)
        {
        if(!m_exec_conf->isRoot())
            return 0;

        std::string path;
        int retval = 0;
        std::vector<uint32_t> N(Ntypes);
        path = name + "N";
        std::transform(shape.cbegin(), shape.cend(), N.begin(), [](const hpmc::detail::poly2d_verts& s) -> uint32_t{return s.N;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_UINT32, Ntypes, 1, 0, (void *)&N[0]);
        path = name + "vertices";
        std::vector<float> data(hpmc::detail::MAX_POLY2D_VERTS*Ntypes*2), sr(Ntypes); // over allocate is ok because we just wont write those extra ones
        uint32_t count = 0;
        for(unsigned int i = 0; i < Ntypes; i++)
            {
            for (unsigned int v = 0; v < shape[i].N; v++)
                {
                data[v*2+0+count] = float(shape[i].x[v]);
                data[v*2+1+count] = float(shape[i].y[v]);
                }
            count += shape[i].N;
            }
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, count, 2, 0, (void *)&data[0]);
        path = name + "sweep_radius";
        std::transform(shape.cbegin(), shape.cend(), sr.begin(), [](const hpmc::detail::poly2d_verts& s) -> float{return s.sweep_radius;});
        retval |= gsd_write_chunk(&handle, path.c_str(), GSD_TYPE_FLOAT, Ntypes, 1, 0, (void *)&sr[0]);
        return retval;
        }

    bool read(  std::shared_ptr<GSDReader> reader,
                uint64_t frame,
                const std::string& name,
                unsigned int Ntypes,
                param_array<hpmc::detail::poly2d_verts>& shape
            )
        {
        bool success = true;
        std::vector<float> vertices,sweep_radius;
        std::vector<uint32_t> N;
        uint32_t count = 0;
        std::string path;
        if(m_exec_conf->isRoot())
            {
            N.resize(Ntypes);
            vertices.resize(hpmc::detail::MAX_POLY2D_VERTS*Ntypes*2);
            sweep_radius.resize(Ntypes);
            path = name + "N";
            success = reader->readChunk((void *)&N[0], frame, path.c_str(),  Ntypes*gsd_sizeof_type(GSD_TYPE_UINT32), Ntypes) && success;
            count = std::accumulate(N.begin(), N.end(), 0);
            path = name + "vertices";
            success = reader->readChunk((void *)&vertices[0], frame, path.c_str(), 2*count*gsd_sizeof_type(GSD_TYPE_FLOAT), count) && success;
            path = name + "sweep_radius";
            success = reader->readChunk((void *)&sweep_radius[0], frame, path.c_str(), Ntypes*gsd_sizeof_type(GSD_TYPE_FLOAT), Ntypes) && success;
            }
        #ifdef ENABLE_MPI
            if(m_mpi)
                {
                bcast(N, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                bcast(vertices, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                bcast(sweep_radius, 0, m_exec_conf->getMPICommunicator()); // broadcast the data
                }
        #endif
            if(!N.size() || !vertices.size() || !sweep_radius.size()) // adding this sanity check but can remove.
                throw std::runtime_error("Error occured while attempting to restore from gsd file.");

        count = 0;
        for (unsigned int i = 0; i < Ntypes; i++)
            {
            float dsq = 0.0;
            for (unsigned int v = 0; v < N[i]; v++)
                {
                shape[i].x[v] = vertices[v*2+0+count];
                shape[i].y[v] = vertices[v*2+1+count];
                dsq = fmax(shape[i].x[v]*shape[i].x[v] + shape[i].y[v]*shape[i].y[v], dsq);
                }
            shape[i].diameter = 2.0*(sqrt(dsq)+shape[i].sweep_radius);
            shape[i].N = N[i];
            shape[i].sweep_radius = sweep_radius[i];
            count += N[i];
            }
        return success;
        }
    };

#endif
